﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Sparkle.Effects.Renderer;
namespace Sparkle.Effects.TextEffects
{


    public abstract class TextEffectObject
    {
        #region DEBUG CODE ONLY - REMOVABLE
        int debugOffset = -100;
        #endregion
        protected internal enum TextEffectState
        {
            Stop,
            In,
            PlayEffect,
            Pause,
            Out,
            Finished
        }


        #region Effect Timers

        /// <summary>
        /// Time in milliseconds of the transition for the effect to appear
        /// </summary>
        protected internal float TransateInDuration=0;
        protected float Elapsed_TransateInDuration;

        /// <summary>
        /// Time in milliseconds of the duration for the effect to play the effect
        /// </summary>
        protected internal float PlayEffectDuration=0;
        protected float Elapsed_PlayEffectDuration;

        
        protected internal  float PauseTime = 0;
        protected float Elapsed_PauseTime;
        /// <summary>
        /// Time in milliseconds of the transition for the effect to disappear,   -1 will keep the text visible at all time
        /// </summary>
        public float FadeOutTimer = 0;
        protected float Elapsed_TransateOutDuration;

        #endregion

        #region Private properties
        /// <summary>
        /// The TextEffectRenderer this effect belongs to
        /// </summary>
        protected internal TextEffectRenderer BaseRenderer;
        /// <summary>
        /// Counter to keep track of the current Block Index
        /// </summary>
        protected internal int BlockIndex = 0;
        /// <summary>
        /// Current transition state of the effect
        /// </summary>
        protected internal TextEffectState TransitionState;
        /// <summary>
        /// List of all TextBlockObjects for the effect
        /// </summary>
        protected internal List<TextBlockObject> Objects = new List<TextBlockObject>();

        /// <summary>
        /// A flag inidcating if the effect has been initialized
        /// </summary>
        private bool IsEffectInitialized = false;
        #endregion

        #region Public Properties
        
        /// <summary>
        /// The font to use when drawing the effect
        /// </summary>
        public SpriteFont Font;
        
        /// <summary>
        /// The position of the effect
        /// </summary>
        public Vector2 Position;
        
        /// <summary>
        /// The text color of the effect
        /// </summary>
        public Color EffectColor = Color.White;
       
        #endregion

        #region Effect events
        public delegate void EffectFinishedEventHandler(TextEffectObject sender);
        public event EffectFinishedEventHandler OnEffectFinished;
        internal void EffectFinishedEvent()
        {
            EffectFinishedEventHandler eventH = OnEffectFinished;
            if (eventH != null)
                OnEffectFinished(this);
        }
        #endregion

        #region Abstract functions that must be implemented by any class using this class as a base class
        public abstract void InitializeEffect(string Text);
        
        
        internal abstract void PlayTransitionOnEffect(float percentElapsed, GameTime gt);
        internal abstract void PlayEffect(float percentElapsed, GameTime gt);
        internal abstract void PlayTransitionOffEffect(float percentElapsed, GameTime gt);

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="renderer"></param>
        public TextEffectObject(TextEffectRenderer renderer)
        {
            BaseRenderer =renderer;
            this.TransitionState = TextEffectState.Stop;
            renderer.AddEffect(this);
        }
        
        /// <summary>Initialize the effect with the selected text and creates the List with TextBlockObjects.
        /// </summary>
        /// <param name="Text">Text for the effect to display</param>
        protected internal void Initialize(string Text, bool StartTransparent)
        {
            
            Objects = new List<TextBlockObject>();
            float MarkerPosition = 0;
            for (int i = 0; i < Text.ToCharArray().Length; i++)
            {
               
               // +Objects[i - 1].BlockPosition.X;
                TextBlockObject obj = new TextBlockObject()
                {
                    BlockColor =StartTransparent? EffectColor*0:EffectColor,
                    BlockOriginalColor = EffectColor,
                    BlockChar = Text.ToCharArray()[i].ToString(),
                    BlockPosition = new Vector2(MarkerPosition, 0),
                    BlockRotation = 0f,
                    BlockScale = 1f,
                    BlockSize = new Vector2(Font.MeasureString(Text.ToCharArray()[i].ToString()).X, Font.MeasureString(Text.ToCharArray()[i].ToString()).Y),
                    PositionInWord_X = MarkerPosition
                };
                Objects.Add(obj);
                MarkerPosition += Objects[i].BlockSize.X;
            }
            IsEffectInitialized = true;
        }
        
        /// <summary> Triggers the effect
        /// </summary>
        public virtual void Trigger()
        {
            if (!IsEffectInitialized)
                throw new EngineException("Effect has not been initialized");
            if(TransitionState == TextEffectState.Stop)
                TransitionState = TextEffectState.In;
        }
      
        /// <summary>Draws the current effect
        /// </summary>
        /// <param name="gt">GameTime</param>
        /// <param name="sp">SpriteBatch</param>
        protected internal virtual void Draw(GameTime gt, SpriteBatch sp)
        {
            for (int i = 0; i < Objects.Count; i++)
            {
                sp.DrawString(
                    Font,
                   Objects[i].BlockChar,
                    //Objects[i].BlockPosition,
                   Vector2.Add(Position, Objects[i].BlockPosition),
                   Objects[i].BlockColor,
                   Objects[i].BlockRotation,
                   Objects[i].BlockOrigin,
                   Objects[i].BlockScale,
                   SpriteEffects.None, 1f
                   );
            }
            #region DEBUG CODE ONLY - REMOVABLE
            sp.DrawString(Sparkle.Effects.Renderer.TextEffectRenderer.debugFont, TransitionState.ToString(), new Vector2(Position.X, Position.Y+10), Color.Red*0.5f);
            sp.DrawString(Sparkle.Effects.Renderer.TextEffectRenderer.debugFont, "Pos: " + Position, new Vector2(Position.X, Position.Y + 20), Color.Red*0.5f);
            //sp.DrawString(Sparkle.Effects.Renderer.TextEffectRenderer.debugFont, "El. Play:" + Elapsed_PlayEffectDuration, new Vector2(Position.X, Position.Y + debugOffset + 20), Color.Red);
            //sp.DrawString(Sparkle.Effects.Renderer.TextEffectRenderer.debugFont, "El. Fade:" + Elapsed_TransateOutDuration, new Vector2(Position.X, Position.Y + debugOffset + 30), Color.Red);
            #endregion
        }
      
        /// <summary> Updates the state of the current effect
        /// </summary>
        /// <param name="gt">GameTime</param>
        protected internal virtual void Update(GameTime gt)
        {
            if (TransitionState == TextEffectState.In)
            {
                Elapsed_TransateInDuration += (float)gt.ElapsedGameTime.TotalMilliseconds;
                
                PlayTransitionOnEffect(Elapsed_TransateInDuration / TransateInDuration, gt);

                if (Elapsed_TransateInDuration >= TransateInDuration)
                {
                    TransitionState = TextEffectObject.TextEffectState.PlayEffect;
                    BlockIndex = 0;
                }
            }
            if (TransitionState == TextEffectState.PlayEffect)
            {
                Elapsed_PlayEffectDuration += (float)gt.ElapsedGameTime.TotalMilliseconds;
             
                PlayEffect(Elapsed_PlayEffectDuration / PlayEffectDuration, gt);

                if (Elapsed_PlayEffectDuration >= PlayEffectDuration)
                {
                    if (FadeOutTimer != -1)
                    {
                        TransitionState = TextEffectState.Pause;
                        BlockIndex = 0;
                    }
                }
            }
            if (TransitionState == TextEffectState.Pause)
            {
                Elapsed_PauseTime += (float)gt.ElapsedGameTime.TotalMilliseconds;

                if (Elapsed_PauseTime>= PauseTime)
                    TransitionState = TextEffectState.Out;
            }

            if (TransitionState == TextEffectState.Out)
            {
                Elapsed_TransateOutDuration += (float)gt.ElapsedGameTime.TotalMilliseconds;

                PlayTransitionOffEffect(Elapsed_TransateOutDuration / FadeOutTimer, gt);

                if (Elapsed_TransateOutDuration >= FadeOutTimer)
                    TransitionState = TextEffectState.Finished;
            }
        }
 
        /// <summary> A common out transition that can be used by any TextEffectObject
        /// </summary>
        /// <param name="percentElapsed">Returns the value between 0 and 1 of the transition state</param>
        /// <param name="ByBlock">If true the effect fades block by block</param>
        internal void FadeOutEffect(float percentElapsed, bool FadeOutByBlock)
        {
            if (FadeOutByBlock)
            {
                if (BlockIndex < Objects.Count)
                {
                    Objects[BlockIndex].BlockColor = Objects[BlockIndex].BlockOriginalColor * (1 - percentElapsed);

                    if (percentElapsed >= 1)
                    {
                        BlockIndex++;
                        Elapsed_TransateOutDuration = 0;
                    }
                }
            }
            else
            {
                for(int i=0;i <Objects.Count;i++)
                    Objects[i].BlockColor = Objects[i].BlockOriginalColor * (1 - percentElapsed);
            }
        }
    }
}